from .connection import Module as Connection
from ..peer import Client as Peer

import deskapp
from deskapp.callback import callbacks
import lank.name
from lank.crypto import get_handler as get_crypto

import random


class Module(deskapp.Module):
    def __init__(self, app, label):
        super().__init__(app)
        self.classID = random.random()
        self.label = label
        self.name = f'+ {label}'

        self.pub_key = None
        self.history = None

        callbacks.append({
            'key': deskapp.Keys.ENTER,
            'docs': '',
            'func': self.on_enter,
            'classID': self.classID,
        })

        self.register_module()

    def page(self, panel):
        w = self.max_w - 2

        if self.app.top.node:
            self.scroll_elements = [
                ('[Refresh]', 'Recent history:'
                                if self.history else None),
            ]

            if self.history:
                for item in self.history.items:
                    dt = item._to_datetime_(
                        item.timestamp).astimezone().strftime(
                            '%a, %d %b %Y @ %I:%M %p')

                    if item.name == lank.name.REGISTER:
                        self.scroll_elements.append((
                            None,
                            f'{dt}: Created new key-pair'))

                    elif item.name == lank.name.PEER:
                        msg = f'{dt}: Signed on'
                        if ':' in item.key[item.key.index(':')+1:]:
                            alias = item.key[item.key.index(':')+1:]
                            alias = alias[alias.index(':')+1:]
                            msg += f' as {alias}'
                        msg += f' at {item.address}'
                        self.scroll_elements.append(('[connect]', msg))

                    else:
                        self.scroll_elements.append((
                            None,
                            f'{dt}: Did something'))

        else:
            self.scroll_elements = [ ]
            panel.addstr(2, 2, self.app.tw(w-2, 'Not connected'),
                self.frontend.color_cb)

        for idx, (button, label) in enumerate(self.scroll_elements):
            if button:
                panel.addstr(idx+2, 11-len(button), button,
                    self.frontend.color_rw
                    if idx == self.scroll else self.frontend.chess_white)
            else:
                panel.addstr(idx+2, 2, self.app.tw(11))

            panel.addstr(idx+2, 11,
                self.app.tw(w-11, f' {label}' if label else ''))

    def on_enter(self, *args, **kwargs):
        if not self.app.top.node:
            return

        if self.scroll == 0:
            self.app.top.node.get_history(self.label, self.on_history)

        else:
            item = self.history.items[self.scroll-1]

            if item.name == lank.name.PEER:
                alias = ''
                if ':' in item.key[item.key.index(':')+1:]:
                    alias = item.key[item.key.index(':')+1:]
                    alias = alias[alias.index(':')+1:]

                host = item.address[:item.address.index(':')]
                port = int(item.address[item.address.index(':')+1:])

                peer = Peer(self.app.top, self.pub_key, host, port,
                    verbose=self.app.top.verbose)
                peer.start()

                mod = Connection(self.app, self, item.label, alias,
                    item.address, peer)
                self.app.logic.setup_panel(mod)

                # FIXME hack:
                idx = self.app._menu.index(self)
                if self.app._menu.index(mod) != idx+1:
                    del self.app._menu[self.app._menu.index(mod)]
                    self.app._menu.insert(idx+1, mod)
                    oldpanels = self.app.logic.available_panels
                    newpanels = { mod.name: oldpanels[mod.name]
                                    for mod in self.app._menu }
                    self.app.logic.available_panels = newpanels
                self.app.logic.cur += 1

    def on_history(self, registration, history):
        crypto = get_crypto(registration.version)
        self.pub_key = crypto.load_public_key(registration.key_pair_pem)
        self.history = history




from lank.peer import (get_handler, HELLO_TIMEOUT, GENERAL_TIMEOUT, HELLO,
                       HELLO_SIZE)
from lank.peer.protocol.v2 import *

from threading import Thread, Event
from queue import Queue, Empty
import socket


KEEPALIVE = 90 #FIXME


class Server(Thread):
    def __init__(self, app, label, port, verbose=True):
        super().__init__(name='peer-server')
        self.app = app
        self.verbose = verbose

        self.label = label
        self.port = port
        self.crypto = None #FIXME
        self.priv_key = None #FIXME

        self.sock = None
        self.peers = { }
        self.buffer = bytearray(HELLO_SIZE)
        self.ready = Event()

    def run(self):
        addr = ('', self.port)
        self.print(f'listening at {addr}')

        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.sock.bind(addr)
            self.sock.listen(5)
            self.ready.set()

            while self.sock:
                client, addr = self.sock.accept()
                self.handle(client, addr)

        except ConnectionRefusedError:
            self.error('[REFUSED]', True)

        except socket.timeout:
            self.error('[TIMEOUT]', True)

        except OSError as e:
            self.error(f'[CLOSED] {e}')

        finally:
            self.sock = None
            self.print('peer server exiting')

            for thread in self.peers.values():
                thread.join()

            self.ready.set()
            self.print('peer server finished')

    def handle(self, sock, addr):
        self.print(f' + connection from {addr}')
        sock.settimeout(HELLO_TIMEOUT)

        try:
            read = sock.recv_into(self.buffer)

            if read == HELLO_SIZE:
                if self.buffer == HELLO:
                    sock.settimeout(GENERAL_TIMEOUT)

                    try:
                        #self.print('getting protocol handler')
                        protocol = get_handler(sock, addr,
                            self.crypto, self.priv_key)

                        if protocol:
                            #self.print('got handler')
                            protocol.print = self.print
                            try:
                                def server():
                                    try:
                                        #self.print('running server')
                                        protocol.server(self)
                                    except Exception as e:
                                        self.error(e)
                                    finally:
                                        self.print(f' - finished {addr}')
                                        #del self.peers[protocol]

                                thread = Thread(
                                    name='peer-server-client',
                                    target=server)

                                self.peers[protocol] = thread
                                thread.start()

                            except BrokenPipeError:
                                self.error(f' - closed {addr} [BROKEN PIPE]',
                                    True, False)

                            except ValueError as e:
                                self.error(f' - terminated {addr}' \
                                    + f' [BAD MESSAGE: {e}]', True)

                        else:
                            self.error(f' - closed {addr} [CLIENT ABORT]', True,
                                False)

                    except ValueError as e:
                        self.error(f' - terminated {addr} [PROTOCOL VERSION]',
                            True, False)

                    except timeout:
                        self.error(f' - terminated {addr} [GENERAL TIMEOUT]',
                            True, False)

                else:
                    self.error(f' - terminated {addr} [BAD HELLO]', True, False)

            elif read:
                self.error(f' - terminated {addr} [BAD HELLO]', True, False)

            else:
                self.error(f' - closed {addr} [CLIENT ABORT]', True, False)

        except socket.timeout:
            self.error(f' - terminated {addr} [HELLO TIMEOUT]', True, False)

    def on_peered(self, handler):
        from .page.peer import Module as Peer
        from .page.connection import Module as Connection

        label = handler.label
        addr = handler.addr
        #pub_key = handler.pub_key

        self.app._notify_(label, f'new connection from {label} {addr}')
        peer = None

        for mod in self.app.desk._menu:
            if isinstance(mod, Peer):
                if mod.label == label:
                    peer = mod
                    break

        if not peer:
            peer = Peer(self.app.desk, label)
            self.app.desk.logic.setup_panel(peer)

        mod = Connection(self.app.desk, peer, label, '', f'{addr[0]}:{addr[1]}',
            handler)
        self.app.desk.logic.setup_panel(mod)

        # FIXME hack:
        idx = self.app.desk._menu.index(peer)
        if self.app.desk._menu.index(mod) != idx+1:
            del self.app.desk._menu[self.app.desk._menu.index(mod)]
            self.app.desk._menu.insert(idx+1, mod)
            oldpanels = self.app.desk.logic.available_panels
            newpanels = { mod.name: oldpanels[mod.name]
                            for mod in self.app.desk._menu }
            self.app.desk.logic.available_panels = newpanels

    def on_text(self, msg, handler):
        self.print(f'server got text: {msg}')

    def stop(self):
        if not self.sock: return
        sock = self.sock
        self.sock = None

        sock.shutdown(socket.SHUT_RDWR)
        sock.close()
        #self.output.put_nowait(None)
        #self.input.put_nowait(None)

        for handler in self.peers:
            handler.sock.shutdown(socket.SHUT_RDWR)
            handler.sock.close()

        self.ready.set()

    def get_public_key(self, crypto, label):
        if not self.app.node:
            return None

        queue = Queue(maxsize=1)

        def on_registration(msg):
            queue.put_nowait(msg)

        self.app.node.get_registration(label, on_registration)
        msg = queue.get()

        return crypto.load_public_key(msg.key_pair_pem)

    def print(self, msg):
        if not self.verbose: return
        self.app.desk.print(msg)

    def error(self, e, force=False, stop=True):
        if not force and not self.sock: return
        #print(f'** ERROR ** {e}')
        self.app.error(e, 'peer server')
        if stop: self.stop()


class Client(Thread):
    def __init__(self, app, pub_key, host, port, verbose=True):
        super().__init__(name='peer-client')
        self.app = app
        self.verbose = verbose

        self.pub_key = pub_key
        self.host = host
        self.port = port

        self.peer = None
        self.input = Queue(maxsize=1)

        self.ping = None

        self.sender_thread = Thread(
            name='peer-client-sender', target=self.sender)
        self.receiver_thread = Thread(
            name='peer-client-receiver', target=self.receiver)

    def run(self):
        addr = (self.host, self.port)
        self.print(f'connecting to peer at {addr}')

        try:
            sock = socket.create_connection(addr, timeout=HELLO_TIMEOUT)
            sock.settimeout(HELLO_TIMEOUT)
            self.peer = Handler(sock, addr,
                self.app.server.crypto, self.app.server.priv_key, self.pub_key)
            self.peer.hello()
            self.peer.sock.settimeout(GENERAL_TIMEOUT)
            #self.print('[READY]')
            #self.print(f'crypto: {self.peer.crypto}')
            #self.print(f'priv_key: {self.peer.priv_key}')

        except ConnectionRefusedError:
            self.error(f'[CONNECTION REFUSED] {addr}', True)
            self.peer = None

        except socket.timeout:
            self.error(f'[CONNECTION TIMEOUT] {addr}', True)
            self.peer = None

        if not self.peer: return
        self.receiver_thread.start()
        self.sender_thread.start()

        label = self.app.server.label
        self.ping = Ping(label)
        self.input.put_nowait(self.ping)

    def text(self, txt):
        self.input.put_nowait(Text(txt))

    def on_text(self, msg, handler):
        self.print(f'client got text: {msg}')

    def stop(self):
        if not self.peer: return
        #self.print('peer client stopping')

        peer = self.peer
        self.peer = None

        peer.sock.shutdown(socket.SHUT_RDWR)
        peer.sock.close()
        #self.output.put_nowait(None)
        self.input.put_nowait(None)

        #import traceback
        #for line in traceback.format_stack():
        #    self.print(f'TRACE: {line.strip()}')

    def join(self):
        self.receiver_thread.join()
        self.sender_thread.join()
        super().join()

    def sender(self):
        #self.print('sender')
        try:
            while self.peer:
                #self.print('checking input queue')
                try:
                    #while msg := self.input.get(timeout=KEEPALIVE):
                    msg = self.input.get(timeout=KEEPALIVE)
                    if msg:
                        if not self.peer: break
                        self.send(msg)

                except Empty:
                    if self.ping:
                        return self.error('ping sent but no pong received')

                    if self.peer:
                        self.ping = Ping(self.app.server.label)
                        self.send(self.ping)

        finally:
            #self.print('sender out')
            self.stop()

    def send(self, msg):
        if not self.peer: return
        self.print(f'C    {self.peer.addr} <- {msg}')
        self.peer.send(msg)

    def receiver(self):
        try:
            #while msg := self.recv():
            msg = self.recv()
            while msg:
                if isinstance(msg, Pong):
                    if not self.ping:
                        return self.error(f'received pong without ping: {msg}')

                    if msg.nonce != self.ping.nonce:
                        return self.error(f'ping-pong nonce mismatch')

                    self.ping = None

                else:
                    #self.output.put(msg)
                    self.print(f'unhandled message: {msg}')

                msg = self.recv()

            self.error('lost connection')

        except OSError as e:
            self.error(e)

        except Exception as e:
            self.error(e)

        finally:
            #self.print('receiver out')
            self.stop()

    def recv(self):
        if not self.peer: return None
        msg = self.peer.recv(self)
        if msg: self.print(f'C    {self.peer.addr} -> {msg}')
        return msg

    def print(self, msg):
        if not self.verbose: return
        self.app.desk.print(msg)

    def error(self, e, force=False):
        #txt = f'{e}'
        #if not txt: txt = type(e)
        #self.print(f'** PEER CLIENT ERROR ** {txt}')
        if not force and not self.peer: return
        self.app.error(e, 'peer client')
        self.stop()




from .__version__ import __version__
from . import config
from .page.connect import Module as Connect
from .page.labels import Module as Labels
from .page.peers import Module as Peers
from .page.peer import Module as Peer
from .peer import Server
from .node import Client as Node

import deskapp
import lank.name

from threading import Thread
import sys


class Application:
    def __init__(self):
        print(f'desklank v{__version__}')
        self.header = 'Peer-to-Peer Encrypted Communication: Connect to Begin'
        self.verbose = '-v' in sys.argv or '--verbose' in sys.argv

        try:
            from .systray import Icon as TrayIcon
            self.tray_icon = TrayIcon(self)
            self.tray_thread = Thread(target=self.tray_icon.run)

        except Exception as e:
            em = str(e)
            if not em: em = type(e)
            print(f'WARNING - Unable to setup system tray icon: {em}')
            self.tray_icon = None
            self.tray_thread = None

        self.desk = deskapp.App(
            modules=[Connect, Labels, Peers],
            title=f'desklank v{__version__}',
            header=self.header,
            v_split=(0.35 if self.verbose else 0.18),
            h_split=0.2,
            demo_mode=False)
        self.desk.top = self
        self.desk.tw = self.tw
        self.desk.data['labels'] = [ ]

        self.server = None
        self.node = None
        self.interests = config.load_interests()
        self.peers = { label: None for label in config.load_peers() }

        self.finished = False
        self.exiting = False

    def run(self):
        try:
            if self.tray_thread:
                self.tray_thread.start()

            self.desk.start()
            self.finished = True

        except KeyboardInterrupt:
            pass

        finally:
            self.quit()
            if self.tray_thread:
                self.tray_thread.join()

    def quit(self):
        if self.exiting: return
        self.exiting = True

        if self.tray_icon:
            self.tray_icon.stop()

        if not self.finished:
            self.desk.close()

        self.disconnect()

    def connect(self, port, label, pwd, alias):
        self.desk.print('connecting')
        self.desk.data['connect'] = False
        self.desk.set_header(
            self.tw(self.desk.frontend.screen_w - 2, ''))

        self.server = Server(self, label, port, verbose=self.verbose)
        self.server.start()
        self.server.ready.wait()

        if self.server.sock:
            self.node = Node(self, port, label, pwd, alias,
                self.error, self.on_connect, verbose=self.verbose)
            self.node.start()

        else:
            self.disconnect()

    def disconnect(self):
        if not self.server: return
        self.desk.print('disconnecting')
        self.desk.data['labels'] = [ ]

        server = self.server
        self.server = None
        server.stop()
        server.join()

        if self.node:
            node = self.node
            self.node = None
            node.stop()
            node.join()

        self.desk.print('finished')
        self.desk.data['connect'] = None
        self.desk.set_header(
            self.tw(self.desk.frontend.screen_w - 2, self.header))

    def error(self, e, context=None):
        ctxt = f'<{context}> ' if context else ''
        etxt = f'{e}'
        if not etxt: etxt = type(e)
        self.desk.print(f'** ERROR {ctxt}** {etxt}')

    def on_connect(self, labels):
        self.desk.data['connect'] = True
        self.desk.data['labels'] = sorted(labels)

        for label in self.desk.data['labels']:
            if label in self.peers:
                if not self.peers[label]:
                    page = Peer(self.desk, label)
                    self.desk.logic.setup_panel(page)
                    self.peers[label] = page

        self.desk.print('connected and ready')

    def notify(self, signed):
        if signed.name == lank.name.REGISTER:
            self._notify_(signed.label, f'{signed.label} has changed keys')

        elif signed.name == lank.name.PEER:
            msg = f'{signed.label} signed on'
            if ':' in signed.key[signed.key.index(':')+1:]:
                alias = signed.key[signed.key.index(':')+1:]
                alias = alias[alias.index(':')+1:]
                msg += f' as {alias}'
            self._notify_(signed.label, msg)

        else:
            self._notify_(signed.label, f'{signed.label} has done something')

    def _notify_(self, title, msg):
        if self.tray_icon:
            self.tray_icon.notify(msg, f'{title} | desklank')

        self.desk.set_header(self.tw(self.desk.frontend.screen_w - 2,
            f'[{title}] >> {msg}'))

    def tw(self, w, txt='', c=' '):
        if not isinstance(txt, str): txt = f'{txt}'
        size = len(txt)
        if size >= w: return txt
        return txt + c*(w-size)




from pystray import Icon as Base, Menu, MenuItem
from PIL import Image, ImageDraw


class Icon(Base):
    def __init__(self, app):
        self.app = app
        icon_image = self.create_image(64, 64, 'green', 'white')

        super().__init__(
            'desklank',
            icon=icon_image,
            menu=Menu(
                #MenuItem(
                #    'With submenu',
                #    Menu(
                #        MenuItem(
                #            'Submenu item 1',
                #            lambda icon, item: 1
                #        ),
                #        MenuItem(
                #            'Submenu item 2',
                #            lambda icon, item: 2
                #        )
                #    )
                #),
                #MenuItem('Show / Hide', self.app.show_hide, default=True),
                MenuItem('Quit', self.app.quit)
            )
        )

    @staticmethod
    def create_image(width, height, color1, color2):
        image = Image.new('RGBA', (width, height), color1)
        dc = ImageDraw.Draw(image)
        dc.rectangle(
            (width // 2, 0, width, height // 2),
            fill=color2)
        dc.rectangle(
            (0, height // 2, width // 2, height),
            fill=color2)

        return image




from .. import config

import deskapp

import random


CLASS_ID = random.random()

class Module(deskapp.Module):
    name = 'Interests'

    def __init__(self, app):
        super().__init__(app)
        self.classID = CLASS_ID

        self.register_module()

    def page(self, panel):
        labels = self.app.data['labels']
        w = self.max_w - 2

        if self.app.top.node:
            self.scroll_elements = [
                ('[Refresh]', 'Select the labels you are interested in:'
                                if labels else None),
            ]

            if labels:
                self.scroll_elements.extend(
                    [('[*]' if label in self.app.top.interests else '[ ]',
                    label) for label in labels])

        else:
            self.scroll_elements = [ ]
            panel.addstr(2, 2, self.app.tw(w-2, 'Not connected'),
                self.frontend.color_cb)

        for idx, (button, label) in enumerate(self.scroll_elements):
            panel.addstr(idx+2, 11-len(button), button, self.frontend.color_rw
                if idx == self.scroll else self.frontend.chess_white)

            panel.addstr(idx+2, 11,
                self.app.tw(w-11, f' {label}' if label else ''))

    @deskapp.callback(CLASS_ID, deskapp.Keys.ENTER)
    def on_enter(self, *args, **kwargs):
        if not self.app.top.node:
            return

        if self.scroll == 0:
            self.app.top.node.get_labels(self.on_labels)

        else:
            label = self.app.data['labels'][self.scroll-1]
            if label in self.app.top.interests:
                del self.app.top.interests[self.app.top.interests.index(label)]
                notify = False
            else:
                self.app.top.interests.append(label)
                notify = True
            config.save_interests(self.app.top.interests)
            self.app.top.node.interest(label, notify)

    def on_labels(self, labels):
        self.app.data['labels'] = sorted(labels)




from ..peer import Client as Peer

import deskapp
from deskapp.callback import callbacks
import curses

import random
from datetime import datetime


class Module(deskapp.Module):
    def __init__(self, app, parent, label, alias, address, peer):
        super().__init__(app)
        self.classID = random.random()
        self.parent = parent
        self.label = label
        self.alias = alias
        self.address = address
        self.peer = peer

        #name = '    '
        #if self.alias: name += f'{self.alias} '
        #self.name = f'{name}@ {self.address}'
        if self.alias: name = f' @ {self.alias}'
        else: name = f' @ {self.address}'
        self.name = name

        c = 2
        while name in self.app.logic.available_panels:
            name = f'{self.name} ({c})'
            c += 1
        self.name = name

        self.elements = [
            '[chat]',
            '[file]',
            '[close]',
        ]
        self.el_width = 0
        for el in self.elements:
            self.el_width += len(el)

        callbacks.append({
            'key': deskapp.Keys.ENTER,
            'docs': '',
            'func': self.on_enter,
            'classID': self.classID,
        })

        '''
        callbacks.append({
            'key': deskapp.Keys.UP,
            'docs': '',
            'func': self.on_scroll_up,
            'classID': self.classID,
        })

        callbacks.append({
            'key': deskapp.Keys.DOWN,
            'docs': '',
            'func': self.on_scroll_down,
            'classID': self.classID,
        })
        '''

        callbacks.append({
            'key': curses.KEY_HOME,
            'docs': '',
            'func': self.on_scroll_top,
            'classID': self.classID,
        })

        callbacks.append({
            'key': curses.KEY_END,
            'docs': '',
            'func': self.on_scroll_bottom,
            'classID': self.classID,
        })

        self.max_w -= 2
        self.max_h -= 3

        '''
        self.lines = [f'line {i}' for i in range(100)] #FIXME
        if len(self.lines) > self.max_h:
            self.scroll_elements = [
                None for i in range(len(self.lines) - self.max_h + 1)
            ]
            self.scroll = len(self.scroll_elements) - 1
        else:
            self.scroll_elements = [ None ]
        self.sdiff = len(self.lines) - len(self.scroll_elements)
        '''
        self.lines = [ ]
        self.scroll_elements = [ None ]
        self.sdiff = -1

        self.last_dt = None

        self.register_module()

    def add_history(self, dt, is_self, txt):
        if not (self.last_dt and self.last_dt.date() == dt.date()):
            self.lines.append([dt.strftime('%A, %B %d, %Y'), None, None,
                is_self])

        self.last_dt = dt

        time = f'{int(dt.strftime("%I"))}:{dt.strftime("%M %p")}'
        user = self.label
        if self.alias: user += '/' + self.alias
        user = f'[{user}]'

        rows = [ ]
        pad = len(time) + len(user) + 2
        avail = self.max_w - pad

        if pad > self.max_w - 10:
            raise ValueError('screen width too small: {self.max_w}')

        while len(txt) > avail:
            try:
                space = txt.rindex(' ', 0, avail)
                rows.append(txt[:space])
                txt = txt[space+1:]

            except ValueError:
                rows.append(txt[:avail])
                txt = txt[avail:]

        rows.append(txt)

        self.lines.append([time, user, rows[0], is_self])
        for row in rows[1:]:
            self.lines.append([len(time), len(user), row, is_self])

        if len(self.lines) > self.max_h:
            self.scroll_elements = [
                None for i in range(len(self.lines) - self.max_h + 1)
            ]
            self.scroll = len(self.scroll_elements) - 1

        self.sdiff = len(self.lines) - len(self.scroll_elements)

    def page(self, panel):
        w = int((self.max_w - 2 - self.el_width) / (len(self.elements) + 1))
        h = 1

        for idx, el in enumerate(self.elements):
            color = self.frontend.color_rw if idx == self.cur_el \
                else self.frontend.chess_white

            panel.addstr(1, h + w, el, color)
            h += w + len(el)

        dt_color = self.frontend.chess_white
        self_color = self.frontend.color_gb
        user_color = self.frontend.color_cb

        for v in range(self.max_h):
            y = self.max_h + 1 - v
            idx = self.scroll + self.sdiff - v
            if idx < 0: break

            line = self.lines[idx]
            dt = line[0]
            user = line[1]
            txt = line[2]
            is_self = line[3]

            if isinstance(user, str) and txt:
                panel.addstr(y, 1, dt, dt_color)
                panel.addstr(y, 1 + len(dt), ' ')
                panel.addstr(y, 2 + len(dt), user,
                    self_color if is_self else user_color)
                panel.addstr(y, 2 + len(dt) + len(user), self.app.tw(
                    self.max_w - len(dt) - len(user) - 1, f' {txt}'))

            elif txt:
                panel.addstr(y, 1, self.app.tw(dt + user + 2))
                panel.addstr(y, 3 + dt + user, self.app.tw(
                    self.max_w - dt - user - 2, txt))

            else:
                panel.addstr(y, 1, self.app.tw(self.max_w, dt), dt_color)

    def on_enter(self, *args, **kwargs):
        if self.cur_el == 2:
            # FIXME hack:
            idx = self.app.logic.cur - 1
            while self.app._menu[idx] is not self.parent:
                idx -= 1
            self.app.logic.cur = idx
            del self.app.logic.available_panels[self.name]
            del self.app._menu[self.app._menu.index(self)]

            self.peer.stop()
            self.peer.join()
            self.app.print('all done with connection')

    def string_decider(self, panel, txt):
        if self.cur_el == 0:
            if txt:
                self.add_history(datetime.now(), True, txt)
                self.peer.text(txt)

    '''
    #def on_scroll_up(self, *args, **kwargs):
    def on_up(self, *args, **kwargs):
        self.app.print('up')

    #def on_scroll_down(self, *args, **kwargs):
    def on_down(self, *args, **kwargs):
        self.app.print('down')
    '''

    def on_scroll_top(self, *args, **kwargs):
        self.scroll = 0

    def on_scroll_bottom(self, *args, **kwargs):
        self.scroll = len(self.scroll_elements) - 1




import setuptools


def run():
    setuptools.setup(
        name='desklank',
        version = findversion('.', 'desklank'),
        author='LANNOCC (Shawn A. Wilson)',
        author_email='lannocc@yahoo.com',
        url='https://github.com/lannocc/desklank',
        packages=setuptools.find_packages(),
        description='Text-mode peer-to-peer encrypted communication utilizing' \
            + ' LANK protocol and DeskApp framework',
        long_description_content_type='text/markdown',
        long_description=open('README.md').read(),
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        install_requires=[
            'appdirs',
            'deskapp >= 0.0.4',
            'lank >= 0.4.7',
            'pillow',
            'pystray',
            'requests',
        ],
        entry_points = {
            'console_scripts': [
                'desklank=desklank.__main__:run',
            ],
        },
        include_package_data=True,
        #data_files = [
        #    ('share/applications', ['com.lannocc.desklank.desktop']),
        #],
    )


def findversion(root, name):
    '''versioning strategy taken from
       http://stackoverflow.com/a/7071358/7203060'''

    import re
    from os.path import join

    vfile = join(root, name, "__version__.py")
    vmatch = re.search(r'^__version__ *= *["\']([^"\']*)["\']',
        open(vfile, "rt").read(), re.M)

    if vmatch:
        version = vmatch.group(1)
        print ("Found %s version %s" % (name, version))
        return version

    else:
        raise RuntimeError("Expecting a version string in %s." % (vfile))


if __name__ == '__main__':
    run()




from .. import config

import deskapp
from lank.peer import DEFAULT_PORT

import random


CLASS_ID = random.random()

class Module(deskapp.Module):
    name = 'Connect'

    def __init__(self, app):
        super().__init__(app)
        self.classID = CLASS_ID

        self.port = DEFAULT_PORT
        self.alias = config.load_connect_alias() or ''
        self.label = config.load_connect_label() or 'anonymous'
        self.pwd = ''
        self.app.data['connect'] = None

        self.register_module()

    def page(self, panel):
        pwd_masked = '*' * len(self.pwd)
        connect = self.app.data['connect']
        connect = 'Disconnect' if connect else 'Cancel' \
                    if connect == False else 'Connect'
        #w = self.max_w - 2
        w = 22

        self.scroll_elements = [
            (None, self.app.tw(12, f'[{connect}]')),
            ('Local Port:', self.app.tw(w, self.port)),
            ('     Alias:', self.app.tw(w, self.alias)),
            ('     Label:', self.app.tw(w, self.label)),
            ('  Password:', self.app.tw(w, pwd_masked)),
        ]

        for idx, (label, value) in enumerate(self.scroll_elements):
            if label:
                panel.addstr(idx+2, 2, label)

            panel.addstr(idx+2, 14, value, self.frontend.color_rw
                if idx == self.scroll else self.frontend.chess_white)

    def string_decider(self, panel, txt):
        if self.scroll == 1:
            try: self.port = int(txt)
            except ValueError: pass
        elif self.scroll == 2:
            self.alias = txt
            self.app.print(self.alias)
        elif self.scroll == 3:
            self.label = txt
        elif self.scroll == 4:
            self.pwd = txt

    @deskapp.callback(CLASS_ID, deskapp.Keys.ENTER)
    def on_enter(self, *args, **kwargs):
        if self.scroll == 0:
            if self.app.data['connect'] is None:
                self.app.top.connect(self.port, self.label, self.pwd,
                                     self.alias)

            else:
                self.app.top.disconnect()




from appdirs import AppDirs

from os.path import join, exists
from os import makedirs


DIRS = AppDirs('desklank', 'LANNOCC')

if not exists(DIRS.user_data_dir):
    makedirs(DIRS.user_data_dir)

def _load_(filename):
    config = join(DIRS.user_data_dir, filename)
    if not exists(config):
        return None

    with open(config, 'r') as config:
        return config.read()

def _save_(filename, txt):
    config = join(DIRS.user_data_dir, filename)

    with open(config, 'w') as config:
        config.write(txt or '')

def load_connect_label():
    return _load_('label')

def save_connect_label(label):
    _save_('label', label)

def load_connect_alias():
    return _load_('alias')

def save_connect_alias(alias):
    _save_('alias', alias)

def load_interests():
    config = _load_('interests')
    if config is None: return [ ]
    return config.split('\n')

def save_interests(interests):
    _save_('interests', '\n'.join(interests))

def load_peers():
    config = _load_('peers')
    if config is None: return [ ]
    return config.split('\n')

def save_peers(peers):
    _save_('peers', '\n'.join(peers))

